package org.me4se.impl.jam;

import java.util.zip.*;
import java.util.*;
import java.io.*;
import java.net.*;

/**
 * Provides a simple ClassLoader that retrieves classes or resources from
 * a JAR or ZIP file.
 * 
 * @author Jrg Pleumann (joerg@pleumann.de)
 */
public class JarClassLoader extends ClassLoader {

	/**
	 * The JAR file on which this ClassLoader operates.
	 */
	private ZipFile jarFile;

	/**
	 * The cache of classes that have already been loaded.
	 */
	private Hashtable classes = new Hashtable();

	/**
	 * Sets the JAR file that the ClassLoader operates on.
	 */
	public void setJarFile(String fileName) {
		try {
			if (jarFile != null) {
				jarFile.close();
				jarFile = null;
			}

            // Don't clear classes for the moment, because loading an existing
            // class a second time results in exception (occurs when starting
            // MIDlet twice)
			//classes.clear();

			if (fileName != null) {
				jarFile = new ZipFile(fileName);
			}
		}
		catch (IOException ignored) {
		}
	}

	/**
	 * Attempts to load the class with the given name. The method first tries
	 * to load the class using the system ClassLoader. If that fails, it tries
	 * to load the class from the JAR file. Note that classes with a package
	 * name of "java.*" can only be loaded using the system ClassLoader due to
	 * security considerations.
	 */
	public Class loadClass(String name, boolean resolveIt) throws ClassNotFoundException {
		// Consult the cache.
		Class c = (Class) classes.get(name);
		if (c != null)
			return c;

        if (!name.startsWith("java.")) {        
            // Load the class from the JAR file
            byte[] b = loadClassData(name);
            if (b != null)
                c = defineClass(name, b, 0, b.length);
        }

        if (c == null) c = findSystemClass(name);

		classes.put(name, c);

		// If desired, resolve all references
		if (resolveIt)
			resolveClass(c);

		return c;
	}

	/**
	 * Accesses a resource. This methods attempts to open an InputStream for
	 * the given resource, first using the system ClassLoader, and, if that
	 * fails, using the JAR file that has been set for this ClassLoader.
	 */
	public InputStream getResourceAsStream(String name) {
		System.out.println("JarClassLoader loading resource: " + name);

		// ZIP names are always relative to the root directory
		if (name.startsWith("/"))
			name = name.substring(1, name.length());

		// Try to get a ZIP file entry and open an InputStream for it
		ZipEntry entry = jarFile.getEntry(name);
		if (entry != null) {
			try {
				return jarFile.getInputStream(entry);
			}
			catch (IOException ex) {
				ex.printStackTrace();
				return null;
			}
		}

		return getSystemResourceAsStream(name);
	}

	/**
	 * Loads the binary class data for the given class. Returns null if the
	 * class can't be loaded due to whatever reason.
	 */
	private byte[] loadClassData(String name) {
		System.out.println("JarClassLoader loading class: " + name);

		// Convert package names to directory separators and append
		// the ".class" suffix.
		String realName = name.replace('.', '/');
		realName += ".class";

		// Try to get a ZIP file entry and read its contents into a byte array.
		ZipEntry entry = jarFile.getEntry(realName);
		if (entry != null) {
			try {
				byte[] buffer = new byte[(int) entry.getSize()];
				BufferedInputStream is = new BufferedInputStream(jarFile.getInputStream(entry));
				is.read(buffer);

				return buffer;
			}
			catch (IOException ex) {
				ex.printStackTrace();
			}
		}

		return null;
	}
}
